/*
 * Copyright (c) 1989 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Muuss.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 *      (B S D) P I N G . C
 *
 * Using the InterNet Control Message Protocol (ICMP) "ECHO" facility,
 * measure round-trip-delays and packet loss across network paths.
 *
 * Author -
 *  Mike Muuss
 *  U. S. Army Ballistic Research Laboratory
 *  December, 1983
 *
 * Status -
 *  Public Domain.  Distribution Unlimited.
 * Bugs -
 *  More statistics could always be gathered.
 *  This program has to run SUID to ROOT to access the ICMP socket.
 *
 * Rewritten/adapted for Waterloo/TCP by
 *  G. Vanem 1997
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <malloc.h>
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <limits.h>
#include <process.h>
#include <io.h>

#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <tcp.h>

#if defined(__GNUC__) || defined(__HIGHC__)
#include <unistd.h>
#endif

#if defined(__HIGHC__)
#pragma Offwarn (39)
#pragma Offwarn (572)
#endif

#undef   MAXHOSTNAMELEN
#define  MAXHOSTNAMELEN 80

#define  DEFDATALEN  (64 - ICMP_MINLEN)  /* default data length */
#define  MAXIPLEN    60
#define  MAXICMPLEN  76
#define  MAXPACKET   (1500 - 60 - ICMP_MINLEN)     /* max packet size */
#define  MAXWAIT     10                  /* max seconds to wait for response */
#define  NROUTES     9                   /* number of record route slots */

#define  A(bit)    rcvd_tbl[(bit)>>3]    /* identify byte in array */
#define  B(bit)    (1 << ((bit) & 0x07)) /* identify bit in byte */
#define  SET(bit)  (A(bit) |= B(bit))
#define  CLR(bit)  (A(bit) &= (~B(bit)))
#define  TST(bit)  (A(bit) & B(bit))

/* various options
 */
int options;
#define  F_FLOOD         0x001
#define  F_INTERVAL      0x002
#define  F_NUMERIC       0x004
#define  F_PINGFILLED    0x008
#define  F_QUIET         0x010
#define  F_RROUTE        0x020
#define  F_SO_DEBUG      0x040
#define  F_SO_DONTROUTE  0x080
#define  F_VERBOSE       0x100

/*
 * MAX_DUP_CHK is the number of bits in received table, i.e. the maximum
 * number of received sequence numbers we can keep track of.  Change 128
 * to 8192 for complete accuracy...
 */
#define  MAX_DUP_CHK  (8 * 128)
int  mx_dup_ck = MAX_DUP_CHK;
char rcvd_tbl [MAX_DUP_CHK / 8];

struct sockaddr whereto;     /* who to ping */
int    datalen = DEFDATALEN;
int    s;                    /* socket file descriptor */
int    ident;                /* id to identify our packets */
u_char outpack [MAXPACKET];
char   BSPACE = '\b';        /* characters written for flood */
char   DOT    = '.';
char  *hostname;

/*
 * counters
 */
long npackets;           /* max packets to transmit */
long nreceived;          /* # of packets we got back */
long nrepeats;           /* number of duplicates */
long ntransmitted;       /* sequence # for outbound packets = #sent */
int  interval = 1;       /* interval between packets */

/*
 * timing
 */
int    timing;           /* flag to do timing */
long   tmin = LONG_MAX;  /* minimum round trip time */
long   tmax;             /* maximum round trip time */
u_long tsum;             /* sum of all times, for doing average */

/*
 * Prototypes
 */
char *pr_addr  (u_long);
void  catcher  (void);
void  pinger   (void);
void  finish   (void);
void  tvsub    (struct timeval *out, struct timeval *in);
void  pr_icmph (struct icmp *icp);
void  pr_retip (struct ip *ip);
void  pr_pack  (u_char *buf, int cc, struct sockaddr_in *from);
void  fill     (char *bp, char *patp);
int   in_cksum (u_short *addr, int len);

char *usage = "usage: ping [-Rdfnqrv] [-c count] [-i wait] "
              "[-l preload]\n\t[-p pattern] [-s packetsize] host\n";

void Exit (char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);
  vfprintf (stderr,fmt,args);
  va_end (args);
  exit (1);
}

int main (int argc, char **argv)
{
  struct  timeval      timeout;
  struct  hostent     *hp;
  struct  sockaddr_in *to;
  struct  protoent    *proto;

  int     i, ch, hold, packlen;
  int     preload = 0;
  u_char *datap   = &outpack [ICMP_MINLEN + sizeof(struct timeval)];
  u_char *packet;
  char   *target;
  char    hnamebuf[MAXHOSTNAMELEN];
  char    rspace  [3 + 4*NROUTES + 1];  /* record route space */

  while ((ch = getopt(argc, argv, "Rc:dfh:i:l:np:qrs:v?")) != EOF)
    switch (ch)
    {
      case 'c':
           npackets = atoi (optarg);
           if (npackets <= 0)
              Exit ("ping: bad number of packets to transmit.\n");
           break;
      case 'd':
           options |= F_SO_DEBUG;
           tcp_set_debug_state (2);
           dbug_init();
           break;
      case 'f':
           options |= F_FLOOD;
           setbuf (stdout,NULL);
           break;
      case 'i':    /* wait between sending packets */
           interval = atoi (optarg);
           if (interval <= 0)
              Exit ("ping: bad timing interval.\n");
           options |= F_INTERVAL;
           break;
      case 'l':
           preload = atoi (optarg);
           if (preload < 0)
              Exit ("ping: bad preload value.\n");
           break;
      case 'n':
           options |= F_NUMERIC;
           break;
      case 'p':    /* fill buffer with user pattern */
           options |= F_PINGFILLED;
           fill ((char*)datap, optarg);
           break;
      case 'q':
           options |= F_QUIET;
           break;
      case 'R':
           options |= F_RROUTE;
           break;
      case 'r':
           options |= F_SO_DONTROUTE;
           break;
      case 's':    /* size of packet to send */
           datalen = atoi (optarg);
           if (datalen > MAXPACKET)
              Exit ("ping: packet size too large.\n");
           if (datalen <= 0)
              Exit ("ping: illegal packet size.\n");
           break;
      case 'v':
           options |= F_VERBOSE;
           break;
      case '?':
      default:
           Exit (usage);
    }

  argc -= optind;
  argv += optind;

  if (argc != 1)
     Exit (usage);

  target = *argv;
  sock_init();

  memset (&whereto,0,sizeof(struct sockaddr));
  to = (struct sockaddr_in *)&whereto;
  to->sin_family      = AF_INET;
  to->sin_addr.s_addr = inet_addr (target);
  if (to->sin_addr.s_addr != INADDR_NONE)
     hostname = target;
  else
  {
    hp = gethostbyname (target);
    if (!hp)
       Exit ("ping: unknown host %s\n", target);

    to->sin_family      = hp->h_addrtype;
    to->sin_addr.s_addr = *(u_long*)hp->h_addr;
    strncpy (hnamebuf, hp->h_name, sizeof(hnamebuf)-1);
    hnamebuf [sizeof(hnamebuf)-1] = '\0';
    hostname = hnamebuf;
  }

  if ((options & F_FLOOD) && (options & F_INTERVAL))
     Exit ("ping: -f and -i incompatible options.\n");

  if (datalen >= sizeof(struct timeval))  /* can we time transfer */
     timing = 1;

  packlen = datalen + MAXIPLEN + MAXICMPLEN;
  if ((packet = malloc(packlen)) == NULL)
     Exit ("ping: out of memory.\n");

  if (!(options & F_PINGFILLED))
     for (i = ICMP_MINLEN; i < datalen; ++i)
         *datap++ = i;

  ident = getpid() & 0xFFFF;

  if ((proto = getprotobyname("icmp")) == NULL)
     Exit ("ping: unknown protocol icmp.\n");

  if ((s = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0)
  {
    perror ("ping: socket");
    return (1);
  }

  hold = 1;
  if (options & F_SO_DEBUG)
     setsockopt (s, SOL_SOCKET, SO_DEBUG, (char*)&hold,sizeof(hold));

  if (options & F_SO_DONTROUTE)
     setsockopt (s, SOL_SOCKET, SO_DONTROUTE, (char*)&hold,sizeof(hold));

  /* record route option
   */
  if (options & F_RROUTE)
  {
    rspace[IPOPT_OPTVAL] = IPOPT_RR;
    rspace[IPOPT_OLEN]   = sizeof(rspace)-1;
    rspace[IPOPT_OFFSET] = IPOPT_MINOFF;
    if (setsockopt(s, IPPROTO_IP, IP_OPTIONS, rspace, sizeof(rspace)) < 0)
    {
      perror ("ping: record route");
      return (1);
    }
  }

  /*
   * When pinging the broadcast address, you can get a lot of answers.
   * Doing something so evil is useful if you are trying to stress the
   * ethernet, or just want to fill the arp cache to get some stuff for
   * /etc/ethers.
   */
  hold = 48 * 1024;
  setsockopt (s, SOL_SOCKET, SO_RCVBUF, (char*)&hold, sizeof(hold));

  if (to->sin_family == AF_INET)
       printf ("ping %s (%s): %d data bytes\n", hostname,
               inet_ntoa(to->sin_addr), datalen);
  else printf ("ping %s: %d data bytes\n", hostname, datalen);

  signal (SIGINT,(void(*)(int))finish);

  while (preload--)    /* fire off them quickies */
     pinger();

  if ((options & F_FLOOD) == 0)
     catcher();    /* start things going */

  for (;;)
  {
    struct sockaddr_in from;
    int cc, fromlen;

    if (options & F_FLOOD)
    {
      fd_set fdmask;
      pinger();
      timeout.tv_sec  = 0;
      timeout.tv_usec = 10000;
      FD_ZERO (&fdmask);
      FD_SET (s,&fdmask);
      if (select_s(s+1,&fdmask,NULL,NULL,&timeout) < 1)
         continue;
    }
    fromlen = sizeof(from);
    cc = recvfrom (s,packet,packlen,0,(struct sockaddr*)&from,&fromlen);
    if (cc < 0)
    {
      if (errno == EINTR)
         break;
      perror ("ping: recvfrom");
      continue;
    }
    pr_pack (packet,cc,&from);
    if (npackets && nreceived >= npackets)
       break;
  }
  finish();
  return (0);
}

/*
 * catcher --
 *  This routine causes another PING to be transmitted, and then
 * schedules another SIGALRM for 1 second from now.
 * 
 * bug --
 *  Our sense of time will slowly skew (i.e., packets will not be
 * launched exactly at 1-second intervals).  This does not affect the
 * quality of the delay and loss statistics.
 */
void catcher (void)
{
  int waittime;

  pinger();
  signal (SIGALRM, (void(*)(int))catcher);

  if (!npackets || ntransmitted < npackets)
     alarm (interval);
  else
  {
    if (nreceived)
    {
      waittime = 2 * tmax / 1000;
      if (waittime == 0)
          waittime = 1;
    }
    else
      waittime = MAXWAIT;
    signal (SIGALRM, (void(*)(int))finish);
    alarm (waittime);
  }
}

/*
 * pinger --
 *   Compose and transmit an ICMP ECHO REQUEST packet.  The IP packet
 * will be added on by the kernel.  The ID field is our UNIX process ID,
 * and the sequence number is an ascending integer.  The first 8 bytes
 * of the data portion are used to hold a UNIX "timeval" struct in VAX
 * byte-order, to compute the round-trip time.
 */
void pinger (void)
{
  struct icmp *icp;
  int    cc, i;

  icp = (struct icmp *)outpack;
  icp->icmp_type  = ICMP_ECHO;
  icp->icmp_code  = 0;
  icp->icmp_cksum = 0;
  icp->icmp_seq   = ntransmitted++;
  icp->icmp_id    = ident;      /* ID */

  CLR (icp->icmp_seq % mx_dup_ck);

  if (timing)
     gettimeofday ((struct timeval*)&outpack[8],NULL);

  cc = datalen + ICMP_MINLEN;      /* skips ICMP portion */

  /* compute ICMP checksum here
   */
  icp->icmp_cksum = in_cksum ((u_short*)icp, cc);

  i = sendto (s,(char*)outpack,cc,0,&whereto,sizeof(struct sockaddr));

  if (i < 0 || i != cc)
  {
    if (i < 0)
       perror ("ping: sendto");
    printf ("ping: wrote %s %d chars, ret=%d\n", hostname, cc, i);
  }
  if (!(options & F_QUIET) && (options & F_FLOOD))
     write (fileno(stdout), &DOT, 1);
}

/*
 * pr_pack --
 *  Print out the packet, if it came from us.  This logic is necessary
 * because ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  This permits multiple copies of this
 * program to be run without having intermingled output (or statistics!).
 */
void pr_pack (u_char *buf, int cc, struct sockaddr_in *from)
{
  static int  old_rrlen;
  static char old_rr[MAX_IPOPTLEN];
  struct icmp   *icp;
  struct ip     *ip;
  struct timeval tv, *tp;
  u_long l;
  u_char *cp, *dp;
  long   triptime = 0;
  int    i, j, hlen, dupflag;

  gettimeofday (&tv,NULL);

  /* Check the IP header
   */
  ip   = (struct ip*)buf;
  hlen = ip->ip_hl << 2;
  if (cc < hlen + ICMP_MINLEN)
  {
    if (options & F_VERBOSE)
       fprintf (stderr, "ping: packet too short (%d bytes) from %s\n",
                cc, inet_ntoa(from->sin_addr));
    return;
  }

  /* Now the ICMP part
   */
  cc -= hlen;
  icp = (struct icmp*)(buf + hlen);
  if (icp->icmp_type == ICMP_ECHOREPLY)
  {
    if (icp->icmp_id != ident)
       return;      /* That was not our ECHO */

    ++nreceived;
    if (timing)
    {
      tp = (struct timeval*) icp->icmp_data;
      tvsub (&tv, tp);
      triptime = tv.tv_sec * 1000 + (tv.tv_usec / 1000);
      tsum += triptime;
      if (triptime < tmin)
          tmin = triptime;
      if (triptime > tmax)
          tmax = triptime;
    }

    if (TST(icp->icmp_seq % mx_dup_ck))
    {
      ++nrepeats;
      --nreceived;
      dupflag = 1;
    }
    else
    {
      SET (icp->icmp_seq % mx_dup_ck);
      dupflag = 0;
    }

    if (options & F_QUIET)
       return;

    if (options & F_FLOOD)
       write (fileno(stdout), &BSPACE, 1);
    else
    {
      printf ("%d bytes from %s: icmp_seq=%u ttl=%d", cc,
              inet_ntoa(from->sin_addr), icp->icmp_seq, ip->ip_ttl);
      if (timing)
         printf (" time=%ld ms", triptime);
      if (dupflag)
         printf (" (DUP!)");

      /* check the data
       */
      cp = (u_char*) (icp->icmp_data+8);
      dp = &outpack [ICMP_MINLEN + sizeof(struct timeval)];
      for (i = ICMP_MINLEN; i < datalen; ++i, ++cp, ++dp)
      {
        if (*cp != *dp)
        {
          printf ("\nwrong data byte #%d should be 0x%x but was 0x%x",
                  i, *dp, *cp);
          cp = (u_char*) icp->icmp_data;
          for (i = ICMP_MINLEN; i < datalen; ++i, ++cp)
          {
            if ((i % 32) == 8)
               printf ("\n\t");
            printf ("%x ", *cp);
          }
          break;
        }
      }
    }
  }
  else
  {
    /* We've got something other than an ECHOREPLY
     */
    if (!(options & F_VERBOSE))
       return;

    printf ("%d bytes from %s: ", cc, pr_addr(from->sin_addr.s_addr));
    pr_icmph (icp);
  }

  /* Display any IP options
   */
  cp = buf + sizeof(struct ip);

  for (; hlen > (int)sizeof(struct ip); --hlen, ++cp)
    switch (*cp)
    {
      case IPOPT_EOL:
           hlen = 0;
           break;
      case IPOPT_LSRR:
           printf ("\nLSRR: ");
           hlen -= 2;
           j = *++cp;
           ++cp;
           if (j > IPOPT_MINOFF)
             for (;;)
             {
               GETLONG (l,cp);
               if (l == 0)
                    printf ("\t0.0.0.0");
               else printf ("\t%s", pr_addr(l));
               hlen -= 4;
               j    -= 4;
               cp   += 4;
               if (j <= IPOPT_MINOFF)
                  break;
               putchar ('\n');
             }
           break;
      case IPOPT_RR:
           j = *++cp;    /* get length */
           i = *++cp;    /* and pointer */
           hlen -= 2;
           if (i > j)
               i = j;
           i -= IPOPT_MINOFF;
           if (i <= 0)
              continue;
           if (i == old_rrlen &&
               cp == (u_char *)buf + sizeof(struct ip) + 2 &&
               !memcmp(cp, old_rr, i) && !(options & F_FLOOD))
           {
             printf ("\t(same route)");
             i = ((i + 3) / 4) * 4;
             hlen -= i;
             cp   += i;
             break;
           }
           old_rrlen = i;
           memcpy (old_rr,cp,i);
           printf ("\nRR: ");
           for (;;)
           {
             GETLONG (l,cp);
             if (l == 0)
                  printf ("\t0.0.0.0");
             else printf ("\t%s", pr_addr(l));
             hlen -= 4;
             i    -= 4;
             cp   += 4;
             if (i <= 0)
                break;
             putchar ('\n');
           }
           break;
      case IPOPT_NOP:
           printf ("\nNOP");
           break;
      default:
           printf ("\nunknown option %x", *cp);
           break;
    }

  if (!(options & F_FLOOD))
  {
    putchar ('\n');
    fflush (stdout);
  }
}

/*
 * in_cksum --
 *  Checksum routine for Internet Protocol family headers (C Version)
 */
int in_cksum (u_short *addr, int len)
{
  int     nleft  = len;
  u_short *w     = addr;
  int     sum    = 0;
  u_short answer = 0;

  /*
   * Our algorithm is simple, using a 32 bit accumulator (sum), we add
   * sequential 16 bit words to it, and at the end, fold back all the
   * carry bits from the top 16 bits into the lower 16 bits.
   */
  while (nleft > 1)
  {
    sum += *w++;
    nleft -= 2;
  }

  /* mop up an odd byte, if necessary
   */
  if (nleft == 1)
  {
    *(u_char*)&answer = *(u_char*)w;
    sum += answer;
  }

  /* add back carry outs from top 16 bits to low 16 bits
   */
  sum  = (sum >> 16) + (sum & 0xFFFF);  /* add hi 16 to low 16 */
  sum += (sum >> 16);                   /* add carry */
  answer = ~sum;                        /* truncate to 16 bits */
  return (answer);
}

/*
 * tvsub --
 *  Subtract 2 timeval structs:  out = out - in.  Out is assumed to
 * be >= in.
 */
void tvsub (struct timeval *out, struct timeval *in)
{
  if ((out->tv_usec -= in->tv_usec) < 0)
  {
    out->tv_sec--;
    out->tv_usec += 1000000;
  }
  out->tv_sec -= in->tv_sec;
}

/*
 * finish --
 *  Print out statistics, and give up.
 */
void finish (void)
{
  signal (SIGINT, SIG_IGN);
  putchar ('\n');
  fflush (stdout);
  printf ("--- %s ping statistics ---\n", hostname);
  printf ("%ld packets transmitted, ", ntransmitted);
  printf ("%ld packets received, ", nreceived);
  if (nrepeats)
     printf ("+%ld duplicates, ", nrepeats);

  if (ntransmitted)
  {
    int loss = (int) (((ntransmitted - nreceived) * 100) / ntransmitted);

    if (nreceived > ntransmitted)
         printf ("-- somebody's printing up packets!");
    else printf ("%d%% packet loss", loss);
  }
  putchar ('\n');
  if (nreceived && timing)
     printf ("round-trip min/avg/max = %ld/%lu/%ld ms\n",
             tmin, tsum / (nreceived + nrepeats), tmax);

  close (s);
  exit (0);
}


/*
 * pr_icmph --
 *  Print a descriptive string about an ICMP header.
 */
void pr_icmph (struct icmp *icp)
{
  switch (icp->icmp_type)
  {
    case ICMP_ECHOREPLY:
         printf ("Echo Reply\n");
         /* XXX ID + Seq + Data */
         break;

    case ICMP_UNREACH:
         switch (icp->icmp_code)
         {
           case ICMP_UNREACH_NET:
                printf ("Destination Net Unreachable\n");
                break;
           case ICMP_UNREACH_HOST:
                printf ("Destination Host Unreachable\n");
                break;
           case ICMP_UNREACH_PROTOCOL:
                printf ("Destination Protocol Unreachable\n");
                break;
           case ICMP_UNREACH_PORT:
                printf ("Destination Port Unreachable\n");
                break;
           case ICMP_UNREACH_NEEDFRAG:
                printf ("frag needed and DF set\n");
                break;
           case ICMP_UNREACH_SRCFAIL:
                printf ("Source Route Failed\n");
                break;
           default:
                printf ("Dest Unreachable, Bad Code: %d\n", icp->icmp_code);
                break;
         }
         /* Print returned IP header information */
         pr_retip ((struct ip *)icp->icmp_data);
         break;

    case ICMP_SOURCEQUENCH:
         printf ("Source Quench\n");
         pr_retip ((struct ip *)icp->icmp_data);
         break;

    case ICMP_REDIRECT:
         switch (icp->icmp_code)
         {
           case ICMP_REDIRECT_NET:
                printf ("Redirect Network");
                break;
           case ICMP_REDIRECT_HOST:
                printf ("Redirect Host");
                break;
           case ICMP_REDIRECT_TOSNET:
                printf ("Redirect Type of Service and Network");
                break;
           case ICMP_REDIRECT_TOSHOST:
                printf ("Redirect Type of Service and Host");
                break;
           default:
                printf ("Redirect, Bad Code: %d", icp->icmp_code);
                break;
         }
         printf ("(New addr: 0x%08lx)\n", icp->icmp_gwaddr.s_addr);
         pr_retip ((struct ip *)icp->icmp_data);
         break;

    case ICMP_ECHO:
         printf ("Echo Request\n");
         /* XXX ID + Seq + Data */
         break;

    case ICMP_TIMXCEED:
         switch (icp->icmp_code)
         {
           case ICMP_TIMXCEED_INTRANS:
                printf ("Time to live exceeded\n");
                break;
           case ICMP_TIMXCEED_REASS:
                printf ("Frag reassembly time exceeded\n");
                break;
           default:
                printf ("Time exceeded, Bad Code: %d\n", icp->icmp_code);
                break;
         }
         pr_retip ((struct ip *)icp->icmp_data);
         break;

    case ICMP_PARAMPROB:
         printf ("Parameter problem: pointer = 0x%02x\n", icp->icmp_pptr);
         pr_retip ((struct ip *)icp->icmp_data);
         break;

    case ICMP_TSTAMP:
         printf ("Timestamp\n");
         /* XXX ID + Seq + 3 timestamps */
         break;

    case ICMP_TSTAMPREPLY:
         printf ("Timestamp Reply\n");
         /* XXX ID + Seq + 3 timestamps */
         break;

    case ICMP_IREQ:
         printf ("Information Request\n");
         /* XXX ID + Seq */
         break;

    case ICMP_IREQREPLY:
         printf ("Information Reply\n");
         /* XXX ID + Seq */
         break;

    case ICMP_MASKREQ:
         printf ("Address Mask Request\n");
         break;

    case ICMP_MASKREPLY:
         printf ("Address Mask Reply\n");
         break;

    default:
         printf ("Bad ICMP type: %d\n", icp->icmp_type);
  }
}

/*
 * pr_iph --
 *  Print an IP header with options.
 */
void pr_iph (struct ip *ip)
{
  int    hlen = ip->ip_hl << 2;
  u_char *cp  = (u_char *)ip + 20;    /* point to options */

  printf ("Vr HL TOS  Len   ID Flg  off TTL Pro  cks      Src      Dst Data\n");
  printf (" %1x  %1x  %02x %04x %04x", ip->ip_v, ip->ip_hl, ip->ip_tos, ip->ip_len, ip->ip_id);
  printf ("   %1x %04x", (ip->ip_off & 0xE000) >> 13, ip->ip_off & 0x1FFF);
  printf ("  %02x  %02x %04x", ip->ip_ttl, ip->ip_p, ip->ip_sum);
  printf (" %s ", inet_ntoa(ip->ip_src));
  printf (" %s ", inet_ntoa(ip->ip_dst));

  /* dump and option bytes */
  while (hlen-- > 20)
        printf ("%02x", *cp++);
  putchar ('\n');
}

/*
 * pr_addr --
 *  Return an ascii host address as a dotted quad and optionally with
 * a hostname.
 */
char * pr_addr (u_long l)
{
  struct hostent *hp;
  static char    buf[80];
  char  *adr = inet_ntoa (*(struct in_addr*)&l);

  if ((options & F_NUMERIC) ||
      (hp = gethostbyaddr((char *)&l, 4, AF_INET)) == NULL)
       sprintf (buf, "%s", adr);
  else sprintf (buf, "%s (%s)", hp->h_name, adr);
  return (buf);
}

/*
 * pr_retip --
 *  Dump some info on a returned (via ICMP) IP packet.
 */
void pr_retip (struct ip *ip)
{
  int     hlen  = ip->ip_hl << 2;
  u_char *cp    = (u_char *)ip + hlen;
  u_short sport = _getshort (cp);
  u_short dport = _getshort (cp+2);

  pr_iph (ip);

  if (ip->ip_p == 6)
     printf ("TCP: from port %u, to port %u (decimal)\n", sport,dport);

  if (ip->ip_p == 17)
     printf ("UDP: from port %u, to port %u (decimal)\n", sport,dport);
}

void fill (char *bp, char *patp)
{
  int   ii, jj, kk;
  int   pat[16];
  char *cp;

  for (cp = patp; *cp; cp++)
    if (!isxdigit(*cp))
       Exit ("ping: patterns must be specified as hex digits.\n");

  ii = sscanf (patp, "%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x%2x",
               &pat[0], &pat[1], &pat[2], &pat[3], &pat[4], &pat[5], &pat[6],
               &pat[7], &pat[8], &pat[9], &pat[10],&pat[11],&pat[12],
               &pat[13],&pat[14],&pat[15]);

  if (ii > 0)
  {
    for (kk = 0; kk <= MAXPACKET - (8 + ii); kk += ii)
        for (jj = 0; jj < ii; ++jj)
            bp[jj + kk] = pat[jj];
  }

  if (!(options & F_QUIET))
  {
    printf ("PATTERN: 0x");
    for (jj = 0; jj < ii; ++jj)
        printf ("%02x", bp[jj] & 0xFF);
    printf ("\n");
  }
}


