
/******************************************************************************
  PING - internet diagnostic tool
  Copyright (C) 1991, University of Waterloo
  portions Copyright (C) 1990, National Center for Supercomputer Applications

  This program is free software; you can redistribute it and/or modify
  it, but you may not sell it.

  This program is distributed in the hope that it will be useful,
  but without any warranty; without even the implied warranty of
  merchantability or fitness for a particular purpose.

      Erick Engelke                   or via E-Mail
      Faculty of Engineering
      University of Waterloo          Erick@development.watstar.uwaterloo.ca
      200 University Ave.,
      Waterloo, Ont., Canada
      N2L 3G1

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <conio.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <tcp.h>

#ifdef __BORLANDC__
#pragma warn -rch
#endif

struct PingPattern {
       BYTE  *data;
       int    len;
     };

struct stddev {
       double sum;        /* standard sum        */
       double sqSum;      /* standard square sum */
       DWORD  samples;    /* # of samples taken  */
     };

struct statistics {
       clock_t startTime;
       double  replyTime;
       DWORD   numReplies;
       DWORD   numPings;
       struct  stddev stddev;
     } stat;

char *name;
int   dbg_mode;

extern DWORD pkt_dropped (void);

void TraceOff (void)
{
  ctrace_on = 0;
}

/*-------------------------------------------------------------------*/

void GetPattern (struct PingPattern *ret, char *pat)
{
  int i = 0;

  ret->data = (BYTE*) calloc (1,2048);  /* ~MTU for PPP */
  assert (ret->data);

  while (pat && *pat)
  {
    char *hex = "0123456789abcdef";
    ret->data[i]    = (strchr(hex,*pat++ | ' ') - hex) << 4;
    ret->data[i++] += (strchr(hex,*pat++ | ' ') - hex);
  }
  ret->len = i;

  for (i = 0; i < ret->len; i++)
     printf ("%02X ", ret->data[i]);
  puts ("");
}

/*-------------------------------------------------------------------*/

double GetTimeAverage (void)
{
  return (stat.stddev.sum / (stat.numReplies * (double)CLOCKS_PER_SEC));
}

/*-------------------------------------------------------------------*/

void PutDeltaTime (double delta)
{
  if (delta < 0.0F)  /* clock() wrapped */
      delta = -delta;

  stat.stddev.sum   += delta;
  stat.stddev.sqSum += pow (delta,2.0);
  stat.stddev.samples++;
}

/*-------------------------------------------------------------------*/

double StdDeviation (void)
{
  double rc, n = (double) stat.stddev.samples;

  if (n <= 1.0)
     return (0.0);
  rc = sqrt ((n*stat.stddev.sqSum - pow(stat.stddev.sum,2.0)) / (n*(n-1.0)));
  return (rc / (double)CLOCKS_PER_SEC);
}

/*-----------------------------------------------------------------------*/

int Stats (void)
{
  DWORD drops = pkt_dropped();

  printf ("\nPing Statistics\n"
          "  Sent        : %lu\n"
          "  Received    : %lu\n", stat.numPings, stat.numReplies);

  if (drops > 0UL)
     printf ("  Dropped     : %lu\n", drops);

  if (stat.numPings)
     printf ("  Success     : %.1f %%\n", 100.0*stat.numReplies/stat.numPings);

  if (!stat.numReplies)
     printf ("  There was no response from %s\n", name);
  else
  {
    printf ("  Average RTT : %5.2f sec\n", GetTimeAverage());
    printf ("  Std Dev     : %5.2f sec\n", StdDeviation());
  }
  return (stat.numReplies > 0);
}

/*-----------------------------------------------------------------------*/

void Usage (char *str)
{
  if (str)
     puts (str);
  puts ("ping [-?vdfst] [-c count] [-w wait] [-p pattern] host [size]");
  exit (1);
}

/*-----------------------------------------------------------------------*/

int main (int argc, char **argv)
{
  DWORD  host, new_rcvd;
  DWORD  send_timeout = 0L;
  DWORD  last_rcvd    = 0L;
  double wait         = 1.0;
  int    flood_ping   = 0;
  int    count        = 0;
  int    trace        = 0;
  BYTE   eth [20];
  struct PingPattern pattern = { NULL,0 };
  int    ch;

  while ((ch = getopt(argc, argv, "?vdfstc:w:p:")) != EOF)
     switch (ch)
     {
       case 'v': puts (wattcpVersion());
                 break;
       case 'd': dbg_mode = 1;
                 break;
       case 's': count = 10;
                 break;
       case '?': Usage (NULL);
                 break;
       case 't': trace = 1;
                 break;
       case 'w': if (flood_ping)
                    Usage ("`-w' option is incompatible with `-f' option\n");
                 wait = atof (optarg);
                 if (wait < 0.001)
                    Usage ("wait must be >= 0.001\n");
                 if (!count)
                    count = 2;
                 break;
       case 'f': flood_ping = 1;
                 if (!count)
                    count = INT_MAX;
                 if (wait != 1.0)
                    Usage ("`-f' option is incompatible with `-w' option\n");
                 break;
       case 'c': count = atoi (optarg);
                 if (count < 1)
                    Usage ("illegal count argument\n");
                 break;
       case 'p': GetPattern (&pattern, optarg);
                 break;
       default : Usage (NULL);
     }

  argc -= optind;
  argv += optind;

  if (argc-- < 1 || (name = *argv++) == NULL)
     Usage (NULL);


  if (dbg_mode)
     dbug_init();

  sock_init();

  if (trace)   /* start tracing now (HighC only!) */
  {
    atexit (TraceOff);
    ctrace_on = 1;
  }

  if (argc)
  {
    pattern.len = atoi (*argv);
    if (pattern.len > mtu-(20+12) || pattern.len < 64)
    {
      printf ("Illegal size (%d - %d)\n", 64, mtu-(20+12));
      Usage (NULL);
    }
    if (!pattern.data)
    {
      pattern.data = malloc (pattern.len);
      memset (pattern.data, 'x', pattern.len);
    }
  }

  if ((host = lookup_host(name,NULL)) == 0)
  {
    printf ("Unable to resolve `%s'\n", name);
    return (3);
  }

  if (!_arp_resolve(host,&eth,0))  /* resolve it before timer starts */
  {
    printf ("Cannot get route to host `%s'\n", name);
    return (4);
  }

  if (flood_ping)
     printf ("pinging with %u flood packets\n", count);

  else if (count > 1)
     printf ("pinging %u times once every %gs\n", count, wait);

  clock();

  if (!flood_ping)
     wait *= 1000.0;    /* get # of milli-sec */

  if (!count)
     count = 1;

  if (dbg_mode)
     printf ("ETH -> %02X:%02X:%02X:%02X:%02X:%02X\n",
             eth[0], eth[1], eth[2], eth[3], eth[4], eth[5]);

  while (1)
  {
    if (!send_timeout || chk_timeout(send_timeout))
    {
      send_timeout = set_timeout ((DWORD)wait);
      if (stat.numPings < count)
      {
        stat.numPings++;
        stat.startTime = clock();
        if (!_ping(host, stat.numPings, pattern.data, pattern.len))
           return Stats();
        if (flood_ping)
             putchar ('.');
        else printf ("\nsent PING # %lu ", stat.numPings);
      }
    }

    if (kbhit())
    {
      getch();
      return Stats();
    }

    tcp_tick (NULL);
    if (_chk_ping(host,&new_rcvd) != 0xFFFFFFFFL)
    {
      stat.replyTime = clock() - stat.startTime;
      PutDeltaTime (stat.replyTime);
      stat.numReplies++;
      if (flood_ping)
      {
        if (new_rcvd != last_rcvd + 1)
             printf ("\b!");
        else printf ("\b");
        send_timeout = 0L;  /* resend immediately */
      }
      else
      {
        printf ("PING receipt # %lu%c: response time %.2f seconds",
                stat.numReplies, new_rcvd == last_rcvd + 1 ? ' ' : '!',
                stat.replyTime / (double)CLOCKS_PER_SEC);
      }
      last_rcvd = new_rcvd;
      if (stat.numReplies == count)
         return Stats();
    }
    fflush (stdout);
  }
  return (0);
}


